# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import datetime
import json
import re
import uuid as uuid_mod
from dataclasses import field, dataclass
from enum import Enum
from io import BytesIO
from re import findall
from typing import List, Optional, Tuple

from importlib.resources import files


ENERGYML_NAMESPACES = {
    "eml": "http://www.energistics.org/energyml/data/commonv2",
    "prodml": "http://www.energistics.org/energyml/data/prodmlv2",
    "witsml": "http://www.energistics.org/energyml/data/witsmlv2",
    "resqml": "http://www.energistics.org/energyml/data/resqmlv2",
}
"""
dict of all energyml namespaces
"""  # pylint: disable=W0105

ENERGYML_NAMESPACES_PACKAGE = {
    "eml": ["http://www.energistics.org/energyml/data/commonv2"],
    "prodml": ["http://www.energistics.org/energyml/data/prodmlv2"],
    "witsml": ["http://www.energistics.org/energyml/data/witsmlv2"],
    "resqml": ["http://www.energistics.org/energyml/data/resqmlv2"],
    "opc": [
        "http://schemas.openxmlformats.org/package/2006/content-types",
        "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
    ],
}
"""
dict of all energyml namespace packages
"""  # pylint: disable=W0105

RGX_ENERGYML_MODULE_NAME = r"energyml\.(?P<pkg>.*)\.v(?P<version>(?P<versionNumber>\d+(_\d+)*)(_dev(?P<versionDev>.*))?)\..*"  # pylint: disable=C0301
RGX_PROJECT_VERSION = r"(?P<n0>[\d]+)(.(?P<n1>[\d]+)(.(?P<n2>[\d]+))?)?"

ENERGYML_MODULES_NAMES = ["eml", "prodml", "witsml", "resqml"]

RELATED_MODULES = [
    ["energyml.eml.v2_0.commonv2", "energyml.resqml.v2_0_1.resqmlv2"],
    [
        "energyml.eml.v2_1.commonv2",
        "energyml.prodml.v2_0.prodmlv2",
        "energyml.witsml.v2_0.witsmlv2",
    ],
    ["energyml.eml.v2_2.commonv2", "energyml.resqml.v2_2_dev3.resqmlv2"],
    [
        "energyml.eml.v2_3.commonv2",
        "energyml.resqml.v2_2.resqmlv2",
        "energyml.prodml.v2_2.prodmlv2",
        "energyml.witsml.v2_1.witsmlv2",
    ],
]

RGX_UUID_NO_GRP = r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
RGX_UUID = r"(?P<uuid>" + RGX_UUID_NO_GRP + ")"
RGX_DOMAIN_VERSION = r"(?P<domainVersion>(?P<versionNum>([\d]+[\._])*\d)\s*(?P<dev>dev\s*(?P<devNum>[\d]+))?)"
RGX_DOMAIN_VERSION_FLAT = r"(?P<domainVersion>(?P<versionNumFlat>([\d]+)*\d)\s*(?P<dev>dev\s*(?P<devNum>[\d]+))?)"


# ContentType
RGX_MIME_TYPE_MEDIA = r"(?P<media>application|audio|font|example|image|message|model|multipart|text|video)"
RGX_CT_ENERGYML_DOMAIN = r"(?P<energymlDomain>x-(?P<domain>[\w]+)\+xml)"
RGX_CT_XML_DOMAIN = r"(?P<xmlRawDomain>(x\-)?(?P<xmlDomain>.+)\+xml)"
RGX_CT_TOKEN_VERSION = r"version=" + RGX_DOMAIN_VERSION
RGX_CT_TOKEN_TYPE = r"type=(?P<type>[\w\_]+)"

RGX_CONTENT_TYPE = (
    RGX_MIME_TYPE_MEDIA
    + "/"
    + "(?P<rawDomain>("
    + RGX_CT_ENERGYML_DOMAIN
    + ")|("
    + RGX_CT_XML_DOMAIN
    + r")|([\w-]+\.?)+)"
    + "(;(("
    + RGX_CT_TOKEN_VERSION
    + ")|("
    + RGX_CT_TOKEN_TYPE
    + ")))*"
)
RGX_QUALIFIED_TYPE = r"(?P<domain>[a-zA-Z]+)" + RGX_DOMAIN_VERSION_FLAT + r"\.(?P<type>[\w_]+)"
# =========

RGX_SCHEMA_VERSION = (
    r"(?P<name>[eE]ml|[cC]ommon|[rR]esqml|[wW]itsml|[pP]rodml|[oO]pc)?\s*v?" + RGX_DOMAIN_VERSION + r"\s*$"
)

RGX_ENERGYML_FILE_NAME_OLD = r"(?P<type>[\w]+)_" + RGX_UUID_NO_GRP + r"\.xml$"
RGX_ENERGYML_FILE_NAME_NEW = RGX_UUID_NO_GRP + r"\.(?P<objectVersion>\d+(\.\d+)*)\.xml$"
RGX_ENERGYML_FILE_NAME = rf"^(.*/)?({RGX_ENERGYML_FILE_NAME_OLD})|({RGX_ENERGYML_FILE_NAME_NEW})"

RGX_XML_HEADER = r"^\s*<\?xml(\s+(encoding\s*=\s*\"(?P<encoding>[^\"]+)\"|version\s*=\s*\"(?P<version>[^\"]+)\"|standalone\s*=\s*\"(?P<standalone>[^\"]+)\"))+"  # pylint: disable=C0301

RGX_IDENTIFIER = rf"{RGX_UUID}(.(?P<version>\w+)?)?"


#    __  ______  ____
#   / / / / __ \/  _/
#  / / / / /_/ // /
# / /_/ / _, _// /
# \____/_/ |_/___/

URI_RGX_GRP_DOMAIN = "domain"
URI_RGX_GRP_DOMAIN_VERSION = "domainVersion"
URI_RGX_GRP_UUID = "uuid"
URI_RGX_GRP_DATASPACE = "dataspace"
URI_RGX_GRP_VERSION = "version"
URI_RGX_GRP_OBJECT_TYPE = "objectType"
URI_RGX_GRP_UUID2 = "uuid2"
URI_RGX_GRP_COLLECTION_DOMAIN = "collectionDomain"
URI_RGX_GRP_COLLECTION_DOMAIN_VERSION = "collectionDomainVersion"
URI_RGX_GRP_COLLECTION_TYPE = "collectionType"
URI_RGX_GRP_QUERY = "query"

# Patterns
_URI_RGX_PKG_NAME = "|".join(ENERGYML_NAMESPACES.keys())  # "[a-zA-Z]+\w+" //witsml|resqml|prodml|eml
URI_RGX = (
    r"^eml:\/\/\/(?:dataspace\('(?P<"
    + URI_RGX_GRP_DATASPACE
    + r">[^']*?(?:''[^']*?)*)'\)\/?)?((?P<"
    + URI_RGX_GRP_DOMAIN
    + r">"
    + _URI_RGX_PKG_NAME
    + r")(?P<"
    + URI_RGX_GRP_DOMAIN_VERSION
    + r">[1-9]\d)\.(?P<"
    + URI_RGX_GRP_OBJECT_TYPE
    + r">\w+)(\((?:(?P<"
    + URI_RGX_GRP_UUID
    + r">(uuid=)?"
    + RGX_UUID_NO_GRP
    + r")|uuid=(?P<"
    + URI_RGX_GRP_UUID2
    + r">"
    + RGX_UUID_NO_GRP
    + r"),\s*version='(?P<"
    + URI_RGX_GRP_VERSION
    + r">[^']*?(?:''[^']*?)*)')\))?)?(\/(?P<"
    + URI_RGX_GRP_COLLECTION_DOMAIN
    + r">"
    + _URI_RGX_PKG_NAME
    + r")(?P<"
    + URI_RGX_GRP_COLLECTION_DOMAIN_VERSION
    + r">[1-9]\d)\.(?P<"
    + URI_RGX_GRP_COLLECTION_TYPE
    + r">\w+))?(?:\?(?P<"
    + URI_RGX_GRP_QUERY
    + r">[^#]+))?$"
)

# ================================
RELS_CONTENT_TYPE = "application/vnd.openxmlformats-package.core-properties+xml"
RELS_FOLDER_NAME = "_rels"

primitives = (bool, str, int, float, type(None))

DOT_PATH_ATTRIBUTE = r"(?:(?<=\\)\.|[^\.])+"
DOT_PATH = rf"\.*(?P<first>{DOT_PATH_ATTRIBUTE})(?P<next>(\.(?P<last>{DOT_PATH_ATTRIBUTE}))*)"


class MimeType(Enum):
    """Some mime types"""

    CSV = "text/csv"
    HDF5 = "application/x-hdf5"
    PARQUET = "application/x-parquet"
    PDF = "application/pdf"
    RELS = "application/vnd.openxmlformats-package.relationships+xml"

    def __str__(self):
        return self.value


class EpcExportVersion(Enum):
    """EPC export version."""

    #: Classical export
    CLASSIC = 1
    #: Export with objet path sorted by package (eml/resqml/witsml/prodml)
    EXPANDED = 2


class EPCRelsRelationshipType(Enum):
    """Rels relationship types"""

    #: The object in Target is the destination of the relationship.
    DESTINATION_OBJECT = "destinationObject"
    #: The current object is the source in the relationship with the target object.
    SOURCE_OBJECT = "sourceObject"
    #: The target object is a proxy object for an external data object (HDF5 file).
    ML_TO_EXTERNAL_PART_PROXY = "mlToExternalPartProxy"
    #: The current object is used as a proxy object by the target object.
    EXTERNAL_PART_PROXY_TO_ML = "externalPartProxyToMl"
    #: The target is a resource outside of the EPC package. Note that TargetMode should be "External"
    #: for this relationship.
    EXTERNAL_RESOURCE = "externalResource"
    #: The object in Target is a media representation for the current object. As a guideline, media files
    #: should be stored in a "media" folder in the ROOT of the package.
    DestinationMedia = "destinationMedia"
    #: The current object is a media representation for the object in Target.
    SOURCE_MEDIA = "sourceMedia"
    #: The target is part of a larger data object that has been chunked into several smaller files
    CHUNKED_PART = "chunkedPart"
    #: The core properties
    CORE_PROPERTIES = "core-properties"
    #: /!\ not in the norm
    EXTENDED_CORE_PROPERTIES = "extended-core-properties"

    def get_type(self) -> str:
        if self == EPCRelsRelationshipType.EXTENDED_CORE_PROPERTIES:
            return "http://schemas.f2i-consulting.com/package/2014/relationships/" + str(self.value)
        elif EPCRelsRelationshipType.CORE_PROPERTIES:
            return "http://schemas.openxmlformats.org/package/2006/relationships/metadata/" + str(self.value)
        # elif (
        #          self == EPCRelsRelationshipType.CHUNKED_PART
        #         or  self == EPCRelsRelationshipType.DESTINATION_OBJECT
        #         or  self == EPCRelsRelationshipType.SOURCE_OBJECT
        #         or  self == EPCRelsRelationshipType.ML_TO_EXTERNAL_PART_PROXY
        #         or  self == EPCRelsRelationshipType.EXTERNAL_PART_PROXY_TO_ML
        #         or  self == EPCRelsRelationshipType.EXTERNAL_RESOURCE
        #         or  self == EPCRelsRelationshipType.DestinationMedia
        #         or  self == EPCRelsRelationshipType.SOURCE_MEDIA
        #     ):
        return "http://schemas.energistics.org/package/2012/relationships/" + str(self.value)


@dataclass
class RawFile:
    """A class for a non energyml file to be stored in an EPC file"""

    path: str = field(default="_")
    content: BytesIO = field(default=None)


#     ______                 __  _
#    / ____/_  ______  _____/ /_(_)___  ____  _____
#   / /_  / / / / __ \/ ___/ __/ / __ \/ __ \/ ___/
#  / __/ / /_/ / / / / /__/ /_/ / /_/ / / / (__  )
# /_/    \__,_/_/ /_/\___/\__/_/\____/_/ /_/____/


def snake_case(string: str) -> str:
    """Transform a str into snake case."""
    string = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", string)
    string = re.sub("__([A-Z])", r"_\1", string)
    string = re.sub("([a-z0-9])([A-Z])", r"\1_\2", string)
    return string.lower()


def pascal_case(string: str) -> str:
    """Transform a str into pascal case."""
    return snake_case(string).replace("_", " ").title().replace(" ", "")


def flatten_concatenation(matrix) -> List:
    """
    Flatten a matrix.

    Example :
        [ [a,b,c], [d,e,f], [ [x,y,z], [0] ] ]
        will be translated in: [a, b, c, d, e, f, [x,y,z], [0]]
    :param matrix:
    :return:
    """
    flat_list = []
    for row in matrix:
        flat_list += row
    return flat_list


def parse_content_type(ct: str) -> Optional[re.Match[str]]:
    return re.search(RGX_CONTENT_TYPE, ct)


def parse_qualified_type(ct: str) -> Optional[re.Match[str]]:
    return re.search(RGX_QUALIFIED_TYPE, ct)


def parse_content_or_qualified_type(cqt: str) -> Optional[re.Match[str]]:
    """
    Give a re.Match object (or None if failed).
    You can access to groups like : "domainVersion", "versionNum", "domain", "type"

    :param cqt:
    :return:
    """
    parsed = None
    try:
        parsed = parse_content_type(cqt)
    except:
        pass
    if parsed is None:
        try:
            parsed = parse_qualified_type(cqt)
        except:
            pass

    return parsed


def get_domain_version_from_content_or_qualified_type(cqt: str) -> Optional[str]:
    """
    return a version number like "2.2" or "2.0"

    :param cqt:
    :return:
    """
    try:
        parsed = parse_content_type(cqt)
        return parsed.group("domainVersion")
    except:
        try:
            parsed = parse_qualified_type(cqt)
            return ".".join(parsed.group("domainVersion"))
        except:
            pass
    return None


def split_identifier(identifier: str) -> Tuple[str, Optional[str]]:
    match = re.match(RGX_IDENTIFIER, identifier)
    return (
        match.group(URI_RGX_GRP_UUID),
        match.group(URI_RGX_GRP_VERSION),
    )


def now(time_zone=datetime.timezone.utc) -> float:
    """Return an epoch value"""
    return datetime.datetime.timestamp(datetime.datetime.now(time_zone))


def epoch(time_zone=datetime.timezone.utc) -> int:
    return int(now(time_zone))


def date_to_epoch(date: str) -> int:
    """
    Transform a energyml date into an epoch datetime
    :return: int
    """
    return int(datetime.datetime.fromisoformat(date).timestamp())


def epoch_to_date(
    epoch_value: int,
) -> str:
    date = datetime.datetime.fromtimestamp(epoch_value, datetime.timezone.utc)
    return date.astimezone(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
    # date = datetime.datetime.fromtimestamp(epoch_value, datetime.timezone.utc)
    # return date.astimezone(datetime.timezone(datetime.timedelta(hours=0), "UTC")).strftime('%Y-%m-%dT%H:%M:%SZ')
    # return date.strftime("%Y-%m-%dT%H:%M:%SZ%z")


def gen_uuid() -> str:
    """
    Generate a new uuid.
    :return:
    """
    return str(uuid_mod.uuid4())


def mime_type_to_file_extension(mime_type: str) -> Optional[str]:
    if mime_type is not None:
        mime_type_lw = mime_type.lower()
        if (
            mime_type_lw == "application/x-parquet"
            or mime_type_lw == "application/parquet"
            or mime_type_lw == "application/vnd.apache.parquet"
        ):
            return "parquet"
        elif mime_type_lw == "application/x-hdf5":
            return "h5"
        elif mime_type_lw == "text/csv":
            return "csv"
        elif mime_type_lw == "application/vnd.openxmlformats-package.relationships+xml":
            return "rels"
        elif mime_type_lw == "application/pdf":
            return "pdf"

    return None


def path_next_attribute(dot_path: str) -> Tuple[Optional[str], Optional[str]]:
    _m = re.match(DOT_PATH, dot_path)
    if _m is not None:
        _next = _m.group("next")
        return _m.group("first"), _next if _next is not None and len(_next) > 0 else None
    return None, None


def path_last_attribute(dot_path: str) -> str:
    _m = re.match(DOT_PATH, dot_path)
    if _m is not None:
        return _m.group("last")
    return None


def path_iter(dot_path: str) -> List[str]:
    return findall(DOT_PATH_ATTRIBUTE, dot_path)


def _get_property_kind_dict_path_as_str(file_type: str = "xml") -> str:
    try:
        import energyml.utils.rc as RC
    except:
        try:
            import src.energyml.utils.rc as RC
        except:
            import utils.rc as RC
    return files(RC).joinpath(f"PropertyKindDictionary_v2.3.{file_type.lower()}").read_text(encoding="utf-8")


def get_property_kind_dict_path_as_json() -> str:
    return _get_property_kind_dict_path_as_str("json")


def get_property_kind_dict_path_as_dict() -> dict:
    return json.loads(_get_property_kind_dict_path_as_str("json"))


def get_property_kind_dict_path_as_xml() -> str:
    return _get_property_kind_dict_path_as_str("xml")


if __name__ == "__main__":

    m = re.match(DOT_PATH, ".Citation.Title.Coucou")
    print(m.groups())
    print(m.group("first"))
    print(m.group("last"))
    print(m.group("next"))
    m = re.match(DOT_PATH, ".Citation")
    print(m.groups())
    print(m.group("first"))
    print(m.group("last"))
    print(m.group("next"))

    print(path_next_attribute(".Citation.Title.Coucou"))
    print(path_iter(".Citation.Title.Coucou"))
    print(path_iter(".Citation.Ti\\.*.Coucou"))

    print(URI_RGX)
    print(RGX_UUID_NO_GRP)



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
"""
This module contains utilities to read/write EPC files.
"""

import datetime
import json
import logging
import os
import random
import re
import traceback
import zipfile
from dataclasses import dataclass, field
from io import BytesIO
from typing import List, Any, Union, Dict, Callable, Optional, Tuple

from energyml.opc.opc import (
    CoreProperties,
    Relationships,
    Types,
    Default,
    Relationship,
    Override,
    Created,
    Creator,
    Identifier,
    Keywords1,
    TargetMode,
)
from xsdata.formats.dataclass.models.generics import DerivedElement

from .constants import (
    RELS_CONTENT_TYPE,
    RELS_FOLDER_NAME,
    RGX_DOMAIN_VERSION,
    EpcExportVersion,
    RawFile,
    EPCRelsRelationshipType,
    MimeType,
    split_identifier,
    get_property_kind_dict_path_as_dict,
)
from .data.datasets_io import (
    read_external_dataset_array,
)
from .exception import UnparsableFile
from .introspection import (
    get_class_from_content_type,
    get_obj_type,
    search_attribute_matching_type,
    get_obj_version,
    get_obj_uuid,
    get_object_type_for_file_path_from_class,
    get_content_type_from_class,
    get_direct_dor_list,
    epoch_to_date,
    epoch,
    gen_uuid,
    get_obj_identifier,
    get_class_from_qualified_type,
    copy_attributes,
    get_obj_attribute_class,
    set_attribute_from_path,
    set_attribute_value,
    get_object_attribute,
    get_qualified_type_from_class, get_class_fields,
)
from .manager import get_class_pkg, get_class_pkg_version
from .serialization import (
    serialize_xml,
    read_energyml_xml_str,
    read_energyml_xml_bytes,
    read_energyml_json_str,
    read_energyml_json_bytes,
    JSON_VERSION,
)
from .workspace import EnergymlWorkspace
from .xml import is_energyml_content_type


@dataclass
class Epc(EnergymlWorkspace):
    """
    A class that represent an EPC file content
    """

    # content_type: List[str] = field(
    #     default_factory=list,
    # )

    export_version: EpcExportVersion = field(default=EpcExportVersion.CLASSIC)

    core_props: CoreProperties = field(default=None)

    """ xml files referred in the [Content_Types].xml  """
    energyml_objects: List = field(
        default_factory=list,
    )

    """ Other files content like pdf etc """
    raw_files: List[RawFile] = field(
        default_factory=list,
    )

    """ A list of external files. It can be used to link hdf5 files """
    external_files_path: List[str] = field(
        default_factory=list,
    )

    """ A list of h5 files stored in memory. (Usefull for Cloud services that doesn't work with local files """
    h5_io_files: List[BytesIO] = field(
        default_factory=list,
    )

    """ 
    Additional rels for objects. Key is the object (same than in @energyml_objects) and value is a list of
    RelationShip. This can be used to link an HDF5 to an ExternalPartReference in resqml 2.0.1
    Key is a value returned by @get_obj_identifier
    """
    additional_rels: Dict[str, List[Relationship]] = field(default_factory=lambda: {})

    """
    Epc file path. Used when loaded from a local file or for export
    """
    epc_file_path: Optional[str] = field(default=None)

    def __str__(self):
        return (
            "EPC file ("
            + str(self.export_version)
            + ") "
            + f"{len(self.energyml_objects)} energyml objects and {len(self.raw_files)} other files {[f.path for f in self.raw_files]}"
            # + f"\n{[serialize_json(ar) for ar in self.additional_rels]}"
        )

    def add_file(self, obj: Union[List, bytes, BytesIO, str, RawFile]):
        """
        Add one ore multiple files to the epc file.
        For non energyml file, it is better to use the RawFile class.
        The input can be a single file content, file path, or a list of them
        :param obj:
        :return:
        """
        if isinstance(obj, list):
            for o in obj:
                self.add_file(o)
        elif isinstance(obj, bytes) or isinstance(obj, BytesIO):
            try:
                xml_obj = read_energyml_xml_bytes(obj)
                self.energyml_objects.append(xml_obj)
            except:
                try:
                    if isinstance(obj, BytesIO):
                        obj.seek(0)
                    json_obj = read_energyml_json_bytes(obj, json_version=JSON_VERSION.OSDU_OFFICIAL)
                    self.add_file(json_obj)
                except:
                    # if isinstance(obj, BytesIO):
                    #     obj.seek(0)
                    # self.add_file(RawFile(path=f"pleaseRenameThisFile_{str(random.random())}", content=obj))
                    raise UnparsableFile()
        elif isinstance(obj, RawFile):
            self.raw_files.append(obj)
        elif isinstance(obj, str):
            # Can be a path or a content
            if os.path.exists(obj):
                with open(obj, "rb") as f:
                    file_content = f.read()
                    f_name = os.path.basename(obj)
                    _, f_ext = os.path.splitext(f_name)
                    if f_ext.lower().endswith(".xml") or f_ext.lower().endswith(".json"):
                        try:
                            self.add_file(file_content)
                        except UnparsableFile:
                            self.add_file(RawFile(f_name, BytesIO(file_content)))
                    elif not f_ext.lower().endswith(".rels"):
                        self.add_file(RawFile(f_name, BytesIO(file_content)))
                    else:
                        logging.error(f"Not supported file extension {f_name}")
            else:
                try:
                    xml_obj = read_energyml_xml_str(obj)
                    self.energyml_objects.append(xml_obj)
                except:
                    try:
                        if isinstance(obj, BytesIO):
                            obj.seek(0)
                        json_obj = read_energyml_json_str(obj, json_version=JSON_VERSION.OSDU_OFFICIAL)
                        self.add_file(json_obj)
                    except:
                        if isinstance(obj, BytesIO):
                            obj.seek(0)
                        self.add_file(RawFile(path=f"pleaseRenameThisFile_{str(random.random())}.txt", content=obj))
        elif str(type(obj).__module__).startswith("energyml."):
            # We should test "energyml.(resqml|witsml|prodml|eml|common)" but I didn't to avoid issues if
            # another specific package comes in the future
            self.energyml_objects.append(obj)
        else:
            logging.error(f"unsupported type {str(type(obj))}")

    # EXPORT functions

    def gen_opc_content_type(self) -> Types:
        """
        Generates a :class:`Types` instance and fill it with energyml objects :class:`Override` values
        :return:
        """
        ct = Types()
        rels_default = Default()
        rels_default.content_type = RELS_CONTENT_TYPE
        rels_default.extension = "rels"

        ct.default = [rels_default]

        ct.override = []
        for e_obj in self.energyml_objects:
            ct.override.append(
                Override(
                    content_type=get_content_type_from_class(type(e_obj)),
                    part_name=gen_energyml_object_path(e_obj, self.export_version),
                )
            )

        if self.core_props is not None:
            ct.override.append(
                Override(
                    content_type=get_content_type_from_class(self.core_props),
                    part_name=gen_core_props_path(self.export_version),
                )
            )

        return ct

    def export_file(self, path: Optional[str] = None) -> None:
        """
        Export the epc file. If :param:`path` is None, the epc 'self.epc_file_path' is used
        :param path:
        :return:
        """
        if path is None:
            path = self.epc_file_path
        epc_io = self.export_io()
        with open(path, "wb") as f:
            f.write(epc_io.getbuffer())

    def export_io(self) -> BytesIO:
        """
        Export the epc file into a :class:`BytesIO` instance. The result is an 'in-memory' zip file.
        :return:
        """
        zip_buffer = BytesIO()

        with zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False) as zip_file:
            # CoreProps
            if self.core_props is None:
                self.core_props = CoreProperties(
                    created=Created(any_element=epoch_to_date(epoch())),
                    creator=Creator(any_element="energyml-utils python module (Geosiris)"),
                    identifier=Identifier(any_element=f"urn:uuid:{gen_uuid()}"),
                    keywords=Keywords1(
                        lang="en",
                        content=["generated;Geosiris;python;energyml-utils"],
                    ),
                    version="1.0",
                )

            zip_info_core = zipfile.ZipInfo(
                filename=gen_core_props_path(self.export_version),
                date_time=datetime.datetime.now().timetuple()[:6],
            )
            data = serialize_xml(self.core_props)
            zip_file.writestr(zip_info_core, data)

            #  Energyml objects
            for e_obj in self.energyml_objects:
                e_path = gen_energyml_object_path(e_obj, self.export_version)
                zip_info = zipfile.ZipInfo(
                    filename=e_path,
                    date_time=datetime.datetime.now().timetuple()[:6],
                )
                data = serialize_xml(e_obj)
                zip_file.writestr(zip_info, data)

            # Rels
            for rels_path, rels in self.compute_rels().items():
                zip_info = zipfile.ZipInfo(
                    filename=rels_path,
                    date_time=datetime.datetime.now().timetuple()[:6],
                )
                data = serialize_xml(rels)
                zip_file.writestr(zip_info, data)

            # Other files:
            for raw in self.raw_files:
                zip_info = zipfile.ZipInfo(
                    filename=raw.path,
                    date_time=datetime.datetime.now().timetuple()[:6],
                )
                zip_file.writestr(zip_info, raw.content.read())

            # ContentType
            zip_info_ct = zipfile.ZipInfo(
                filename=get_epc_content_type_path(),
                date_time=datetime.datetime.now().timetuple()[:6],
            )
            data = serialize_xml(self.gen_opc_content_type())
            zip_file.writestr(zip_info_ct, data)

        return zip_buffer

    def compute_rels(self) -> Dict[str, Relationships]:
        """
        Returns a dict containing for each objet, the rels xml file path as key and the RelationShips object as value
        :return:
        """
        dor_relation = get_reverse_dor_list(self.energyml_objects)

        # destObject
        rels = {
            obj_id: [
                Relationship(
                    target=gen_energyml_object_path(target_obj, self.export_version),
                    type_value=EPCRelsRelationshipType.DESTINATION_OBJECT.get_type(),
                    id=f"_{obj_id}_{get_obj_type(target_obj)}_{get_obj_identifier(target_obj)}",
                )
                for target_obj in target_obj_list
            ]
            for obj_id, target_obj_list in dor_relation.items()
        }
        # sourceObject
        for obj in self.energyml_objects:
            obj_id = get_obj_identifier(obj)
            if obj_id not in rels:
                rels[obj_id] = []
            for target_obj in get_direct_dor_list(obj):
                try:
                    rels[obj_id].append(
                        Relationship(
                            target=gen_energyml_object_path(target_obj, self.export_version),
                            type_value=EPCRelsRelationshipType.SOURCE_OBJECT.get_type(),
                            id=f"_{obj_id}_{get_obj_type(target_obj)}_{get_obj_identifier(target_obj)}",
                        )
                    )
                except Exception:
                    logging.error(f'Failed to create rels for "{obj_id}" with target {target_obj}')

        # filtering non-accessible objects from DOR
        rels = {k: v for k, v in rels.items() if self.get_object_by_identifier(k) is not None}

        map_obj_id_to_obj = {get_obj_identifier(obj): obj for obj in self.energyml_objects}

        obj_rels = {
            gen_rels_path(
                energyml_object=map_obj_id_to_obj.get(obj_id),
                export_version=self.export_version,
            ): Relationships(
                relationship=obj_rels + (self.additional_rels[obj_id] if obj_id in self.additional_rels else []),
            )
            for obj_id, obj_rels in rels.items()
        }

        # CoreProps
        if self.core_props is not None:
            obj_rels[gen_rels_path(self.core_props)] = Relationships(
                relationship=[
                    Relationship(
                        target=gen_core_props_path(),
                        type_value=EPCRelsRelationshipType.CORE_PROPERTIES.get_type(),
                        id="CoreProperties",
                    )
                ]
            )

        return obj_rels

    def rels_to_h5_file(self, obj: any, h5_path: str) -> Relationship:
        """
        Creates in the epc file, a Relation (in the object .rels file) to link a h5 external file.
        Usually this function is used to link an ExternalPartReference to a h5 file.
        In practice, the Relation object is added to the "additional_rels" of the current epc file.
        :param obj:
        :param h5_path:
        :return: the Relationship added to the epc.additional_rels dict
        """
        obj_ident = get_obj_identifier(obj)
        if obj_ident not in self.additional_rels:
            self.additional_rels[obj_ident] = []

        rel = Relationship(
            target=h5_path,
            type_value=EPCRelsRelationshipType.EXTERNAL_RESOURCE.get_type(),
            id="Hdf5File",
            target_mode=TargetMode.EXTERNAL.value,
        )
        self.additional_rels[obj_ident].append(rel)
        return rel

    # -- Functions inherited from EnergymlWorkspace

    def get_object_as_dor(self, identifier: str, dor_qualified_type) -> Optional[Any]:
        """
        Search an object by its identifier and returns a DOR
        :param identifier:
        :param dor_qualified_type: the qualified type of the DOR (e.g. resqml22.DataObjectReference)
        :return:
        """
        obj = self.get_object_by_identifier(identifier=identifier)
        # if obj is None:

        return as_dor(obj_or_identifier=obj or identifier, dor_qualified_type=dor_qualified_type)

    def get_object_by_uuid(self, uuid: str) -> List[Any]:
        """
        Search all objects with the uuid :param:`uuid`.
        :param uuid:
        :return:
        """
        return list(filter(lambda o: get_obj_uuid(o) == uuid, self.energyml_objects))

    def get_object_by_identifier(self, identifier: str) -> Optional[Any]:
        """
        Search an object by its identifier.
        :param identifier: given by the function :func:`get_obj_identifier`
        :return:
        """
        for o in self.energyml_objects:
            if get_obj_identifier(o) == identifier:
                return o
        return None

    def get_object(self, uuid: str, object_version: Optional[str]) -> Optional[Any]:
        return self.get_object_by_identifier(f"{uuid}.{object_version or ''}")

    def get_epc_file_folder(self) -> Optional[str]:
        if self.epc_file_path is not None and len(self.epc_file_path) > 0:
            folders_and_name = re.split(r"[\\/]", self.epc_file_path)
            if len(folders_and_name) > 1:
                return "/".join(folders_and_name[:-1])
            else:
                return ""
        return None

    def read_external_array(
        self,
        energyml_array: Any,
        root_obj: Optional[Any] = None,
        path_in_root: Optional[str] = None,
        use_epc_io_h5: bool = True,
    ) -> List[Any]:
        sources = []
        if self is not None and use_epc_io_h5 and self.h5_io_files is not None and len(self.h5_io_files):
            sources = sources + self.h5_io_files

        return read_external_dataset_array(
            energyml_array=energyml_array,
            root_obj=root_obj,
            path_in_root=path_in_root,
            additional_sources=sources,
            epc=self,
        )

    # Class methods

    @classmethod
    def read_file(cls, epc_file_path: str):
        with open(epc_file_path, "rb") as f:
            epc = cls.read_stream(BytesIO(f.read()))
            epc.epc_file_path = epc_file_path
            return epc

    @classmethod
    def read_stream(cls, epc_file_io: BytesIO):  # returns an Epc instance
        """
        :param epc_file_io:
        :return: an :class:`EPC` instance
        """
        try:
            _read_files = []
            obj_list = []
            raw_file_list = []
            additional_rels = {}
            core_props = None
            with zipfile.ZipFile(epc_file_io, "r", zipfile.ZIP_DEFLATED) as epc_file:
                content_type_file_name = get_epc_content_type_path()
                content_type_info = None
                try:
                    content_type_info = epc_file.getinfo(content_type_file_name)
                except KeyError:
                    for info in epc_file.infolist():
                        if info.filename.lower() == content_type_file_name.lower():
                            content_type_info = info
                            break

                _read_files.append(content_type_file_name)

                if content_type_info is None:
                    logging.error(f"No {content_type_file_name} file found")
                else:
                    content_type_obj: Types = read_energyml_xml_bytes(epc_file.read(content_type_file_name))
                    path_to_obj = {}
                    for ov in content_type_obj.override:
                        ov_ct = ov.content_type
                        ov_path = ov.part_name
                        # logging.debug(ov_ct)
                        while ov_path.startswith("/") or ov_path.startswith("\\"):
                            ov_path = ov_path[1:]
                        if is_energyml_content_type(ov_ct):
                            _read_files.append(ov_path)
                            try:
                                ov_obj = read_energyml_xml_bytes(
                                    epc_file.read(ov_path),
                                    get_class_from_content_type(ov_ct),
                                )
                                if isinstance(ov_obj, DerivedElement):
                                    ov_obj = ov_obj.value
                                path_to_obj[ov_path] = ov_obj
                                obj_list.append(ov_obj)
                            except Exception as e:
                                logging.error(traceback.format_exc())
                                logging.error(
                                    f"Epc.@read_stream failed to parse file {ov_path} for content-type: {ov_ct} => {get_class_from_content_type(ov_ct)}\n\n",
                                    get_class_from_content_type(ov_ct),
                                )
                                try:
                                    logging.debug(epc_file.read(ov_path))
                                except:
                                    pass
                                # raise e
                        elif get_class_from_content_type(ov_ct) == CoreProperties:
                            _read_files.append(ov_path)
                            core_props = read_energyml_xml_bytes(epc_file.read(ov_path), CoreProperties)
                            path_to_obj[ov_path] = core_props

                    for f_info in epc_file.infolist():
                        if f_info.filename not in _read_files:
                            _read_files.append(f_info.filename)
                            if not f_info.filename.lower().endswith(".rels"):
                                try:
                                    raw_file_list.append(
                                        RawFile(
                                            path=f_info.filename,
                                            content=BytesIO(epc_file.read(f_info.filename)),
                                        )
                                    )
                                except IOError as e:
                                    logging.error(traceback.format_exc())
                            elif f_info.filename != "_rels/.rels":  # CoreProperties rels file
                                # RELS FILES READING START

                                # logging.debug(f"reading rels {f_info.filename}")
                                (
                                    rels_folder,
                                    rels_file_name,
                                ) = get_file_folder_and_name_from_path(f_info.filename)
                                while rels_folder.endswith("/"):
                                    rels_folder = rels_folder[:-1]
                                obj_folder = rels_folder[: rels_folder.rindex("/") + 1] if "/" in rels_folder else ""
                                obj_file_name = rels_file_name[:-5]  # removing the ".rels"
                                rels_file: Relationships = read_energyml_xml_bytes(
                                    epc_file.read(f_info.filename),
                                    Relationships,
                                )
                                obj_path = obj_folder + obj_file_name
                                if obj_path in path_to_obj:
                                    try:
                                        additional_rels_key = get_obj_identifier(path_to_obj[obj_path])
                                        for rel in rels_file.relationship:
                                            # logging.debug(f"\t\t{rel.type_value}")
                                            if (
                                                rel.type_value != EPCRelsRelationshipType.DESTINATION_OBJECT.get_type()
                                                and rel.type_value != EPCRelsRelationshipType.SOURCE_OBJECT.get_type()
                                                and rel.type_value
                                                != EPCRelsRelationshipType.EXTENDED_CORE_PROPERTIES.get_type()
                                            ):  # not a computable relation
                                                if additional_rels_key not in additional_rels:
                                                    additional_rels[additional_rels_key] = []
                                                additional_rels[additional_rels_key].append(rel)
                                    except AttributeError:
                                        logging.error(traceback.format_exc())
                                        pass  # 'CoreProperties' object has no attribute 'object_version'
                                    except Exception as e:
                                        logging.error(f"Error with obj path {obj_path} {path_to_obj[obj_path]}")
                                        raise e
                                else:
                                    logging.error(
                                        f"xml file '{f_info.filename}' is not associate to any readable object "
                                        f"(or the object type is not supported because"
                                        f" of a lack of a dependency module) "
                                    )

            return Epc(
                energyml_objects=obj_list,
                raw_files=raw_file_list,
                core_props=core_props,
                additional_rels=additional_rels,
            )
        except zipfile.BadZipFile as error:
            logging.error(error)

        return None


#     ______                                      __   ____                 __  _
#    / ____/___  ___  _________ ___  ______ ___  / /  / __/_  ______  _____/ /_(_)___  ____  _____
#   / __/ / __ \/ _ \/ ___/ __ `/ / / / __ `__ \/ /  / /_/ / / / __ \/ ___/ __/ / __ \/ __ \/ ___/
#  / /___/ / / /  __/ /  / /_/ / /_/ / / / / / / /  / __/ /_/ / / / / /__/ /_/ / /_/ / / / (__  )
# /_____/_/ /_/\___/_/   \__, /\__, /_/ /_/ /_/_/  /_/  \__,_/_/ /_/\___/\__/_/\____/_/ /_/____/
#                       /____//____/

"""
PropertyKind list: a list of Pre-defined properties
"""
__CACHE_PROP_KIND_DICT__ = {}


def update_prop_kind_dict_cache():
    prop_kind = get_property_kind_dict_path_as_dict()

    for prop in prop_kind["PropertyKind"]:
        __CACHE_PROP_KIND_DICT__[prop["Uuid"]] = read_energyml_json_str(json.dumps(prop))[0]


def as_dor(obj_or_identifier: Any, dor_qualified_type: str = "eml23.DataObjectReference"):
    """
    Create an DOR from an object to target the latter.
    :param obj_or_identifier:
    :param dor_qualified_type: the qualified type of the DOR (e.g. "eml23.DataObjectReference" is the default value)
    :return:
    """
    dor = None
    if obj_or_identifier is not None:
        if isinstance(obj_or_identifier, str):  # is an identifier
            cls = get_class_from_qualified_type(dor_qualified_type)
            dor = cls()
            if len(__CACHE_PROP_KIND_DICT__) == 0:
                # update the cache to check if it is a
                update_prop_kind_dict_cache()
            try:
                uuid, version = split_identifier(obj_or_identifier)
                if uuid in __CACHE_PROP_KIND_DICT__:
                    return as_dor(__CACHE_PROP_KIND_DICT__[uuid])
                else:
                    set_attribute_from_path(dor, "uuid", uuid)
                    set_attribute_from_path(dor, "ObjectVersion", version)
            except AttributeError:
                logging.error(f"Failed to parse identifier {obj_or_identifier}. DOR will be empty")
        else:
            cls = get_class_from_qualified_type(dor_qualified_type)
            dor = cls()
            if hasattr(dor, "qualified_type"):
                set_attribute_from_path(dor, "qualified_type", get_qualified_type_from_class(obj_or_identifier))
            if hasattr(dor, "content_type"):
                set_attribute_from_path(dor, "content_type", get_content_type_from_class(obj_or_identifier))

            set_attribute_from_path(dor, "uuid", get_object_attribute(obj_or_identifier, "uuid"))
            set_attribute_from_path(dor, "object_version", get_object_attribute(obj_or_identifier, "ObjectVersion"))
            set_attribute_from_path(dor, "title", get_object_attribute(obj_or_identifier, "Citation.Title"))

    return dor


def create_energyml_object(
    content_or_qualified_type: str,
    citation: Optional[Any] = None,
    uuid: Optional[str] = None,
):
    """
    Create an energyml object instance depending on the content-type or qualified-type given in parameter.
    The SchemaVersion is automatically assigned.
    If no citation is given default one will be used.
    If no uuid is given, a random uuid will be used.
    :param content_or_qualified_type:
    :param citation:
    :param uuid:
    :return:
    """
    if citation is None:
        citation = {
            "title": "New_Object",
            "Creation": epoch_to_date(epoch()),
            "LastUpdate": epoch_to_date(epoch()),
            "Format": "energyml-utils",
            "Originator": "energyml-utils python module",
        }
    cls = get_class_from_qualified_type(content_or_qualified_type)
    obj = cls()
    cit = get_obj_attribute_class(cls, "citation")()
    copy_attributes(
        obj_in=citation,
        obj_out=cit,
        only_existing_attributes=True,
        ignore_case=True,
    )
    set_attribute_from_path(obj, "citation", cit)
    set_attribute_value(obj, "uuid", uuid or gen_uuid())
    set_attribute_value(obj, "SchemaVersion", get_class_pkg_version(obj))

    return obj


def create_external_part_reference(
    eml_version: str,
    h5_file_path: str,
    citation: Optional[Any] = None,
    uuid: Optional[str] = None,
):
    """
    Create an EpcExternalPartReference depending on the energyml version (should be ["2.0", "2.1", "2.2"]).
    The MimeType, ExistenceKind and Filename will be automatically filled.
    :param eml_version:
    :param h5_file_path:
    :param citation:
    :param uuid:
    :return:
    """
    version_flat = re.findall(RGX_DOMAIN_VERSION, eml_version)[0][0].replace(".", "").replace("_", "")
    obj = create_energyml_object(
        content_or_qualified_type="eml" + version_flat + ".EpcExternalPartReference",
        citation=citation,
        uuid=uuid,
    )
    set_attribute_value(obj, "MimeType", MimeType.HDF5.value)
    set_attribute_value(obj, "ExistenceKind", "Actual")
    set_attribute_value(obj, "Filename", h5_file_path)

    return obj


def get_reverse_dor_list(obj_list: List[Any], key_func: Callable = get_obj_identifier) -> Dict[str, List[Any]]:
    """
    Compute a dict with 'OBJ_UUID.OBJ_VERSION' as Key, and list of DOR that reference it.
    If the object version is None, key is 'OBJ_UUID.'
    :param obj_list:
    :param key_func: a callable to create the key of the dict from the object instance
    :return: str
    """
    rels = {}
    for obj in obj_list:
        for dor in search_attribute_matching_type(obj, "DataObjectReference", return_self=False):
            key = key_func(dor)
            if key not in rels:
                rels[key] = []
            rels[key] = rels.get(key, []) + [obj]
    return rels


# PATHS


def gen_core_props_path(
    export_version: EpcExportVersion = EpcExportVersion.CLASSIC,
):
    return "docProps/core.xml"


def gen_energyml_object_path(
    energyml_object: Union[str, Any],
    export_version: EpcExportVersion = EpcExportVersion.CLASSIC,
):
    """
    Generate a path to store the :param:`energyml_object` into an epc file (depending on the :param:`export_version`)
    :param energyml_object:
    :param export_version:
    :return:
    """
    if isinstance(energyml_object, str):
        energyml_object = read_energyml_xml_str(energyml_object)

    obj_type = get_object_type_for_file_path_from_class(energyml_object.__class__)

    pkg = get_class_pkg(energyml_object)
    pkg_version = get_class_pkg_version(energyml_object)
    object_version = get_obj_version(energyml_object)
    uuid = get_obj_uuid(energyml_object)

    # if object_version is None:
    #     object_version = "0"

    if export_version == EpcExportVersion.EXPANDED:
        return f"namespace_{pkg}{pkg_version.replace('.', '')}/{uuid}{(('/version_' + object_version) if object_version is not None else '')}/{obj_type}_{uuid}.xml"
    else:
        return obj_type + "_" + uuid + ".xml"


def get_file_folder_and_name_from_path(path: str) -> Tuple[str, str]:
    """
    Returns a tuple (FOLDER_PATH, FILE_NAME)
    :param path:
    :return:
    """
    obj_folder = path[: path.rindex("/") + 1] if "/" in path else ""
    obj_file_name = path[path.rindex("/") + 1 :] if "/" in path else path
    return obj_folder, obj_file_name


def gen_rels_path(
    energyml_object: Any,
    export_version: EpcExportVersion = EpcExportVersion.CLASSIC,
) -> str:
    """
    Generate a path to store the :param:`energyml_object` rels file into an epc file
    (depending on the :param:`export_version`)
    :param energyml_object:
    :param export_version:
    :return:
    """
    if isinstance(energyml_object, CoreProperties):
        return f"{RELS_FOLDER_NAME}/.rels"
    else:
        obj_path = gen_energyml_object_path(energyml_object, export_version)
        obj_folder, obj_file_name = get_file_folder_and_name_from_path(obj_path)
        return f"{obj_folder}{RELS_FOLDER_NAME}/{obj_file_name}.rels"


def get_epc_content_type_path(
    export_version: EpcExportVersion = EpcExportVersion.CLASSIC,
) -> str:
    """
    Generate a path to store the "[Content_Types].xml" file into an epc file
    (depending on the :param:`export_version`)
    :return:
    """
    return "[Content_Types].xml"



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from typing import Optional, Any, List


@dataclass
class EnergymlWorkspace:
    def get_object(self, uuid: str, object_version: Optional[str]) -> Optional[Any]:
        raise NotImplementedError("EnergymlWorkspace.get_object")

    def get_object_by_identifier(self, identifier: str) -> Optional[Any]:
        _tmp = identifier.split(".")
        return self.get_object(_tmp[0], _tmp[1] if len(_tmp) > 1 else None)

    def get_object_by_uuid(self, uuid: str) -> Optional[Any]:
        return self.get_object(uuid, None)

    def read_external_array(
        self,
        energyml_array: Any,
        root_obj: Optional[Any] = None,
        path_in_root: Optional[str] = None,
    ) -> List[Any]:
        raise NotImplementedError("EnergymlWorkspace.get_object")



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Any, Union

from lxml import etree as ETREE  # type: Any

from .constants import *


def get_pkg_from_namespace(namespace: str) -> Optional[str]:
    for (k, v) in ENERGYML_NAMESPACES_PACKAGE.items():
        if namespace in v:
            return k
    return None


def is_energyml_content_type(content_type: str) -> bool:
    ct = parse_content_type(content_type)
    domain = ct.group("domain")
    return domain is not None and domain in ENERGYML_NAMESPACES_PACKAGE.keys()


def get_root_namespace(tree: ETREE.Element) -> str:
    return tree.nsmap[tree.prefix]


def get_class_name_from_xml(tree: ETREE.Element) -> str:
    root_namespace = get_root_namespace(tree)
    pkg = get_pkg_from_namespace(root_namespace)
    if pkg is None:
        logging.error(f"No pkg found for elt {tree}")
    else:
        if pkg == "opc":
            return "energyml.opc.opc." + get_root_type(tree)
        else:
            schema_version = (
                (find_schema_version_in_element(tree) or "")
                .replace(".", "_")
                .replace("-", "_")
            )
            if pkg == "resqml" and schema_version == "2_0":
                schema_version = "2_0_1"

            return (
                "energyml."
                + pkg
                + ".v"
                + schema_version
                + "."
                + root_namespace[root_namespace.rindex("/") + 1:]
                + "."
                + get_root_type(tree)
            )


def get_xml_encoding(xml_content: str) -> Optional[str]:
    try:
        m = re.search(RGX_XML_HEADER, xml_content)
        return m.group("encoding")
    except AttributeError:
        return "utf-8"


def get_tree(xml_content: Union[bytes, str]) -> ETREE.Element:
    xml_bytes = xml_content
    if isinstance(xml_bytes, str):
        # return ETREE.fromstring(xml_content)
        encoding = get_xml_encoding(xml_content)
        xml_bytes = xml_content.encode(
            encoding=encoding.strip().lower()
            if encoding is not None
            else "utf-8"
        )

    return ETREE.parse(BytesIO(xml_bytes)).getroot()


def energyml_xpath(tree: ETREE.Element, xpath: str) -> Optional[list]:
    """A xpath research that knows energyml namespaces"""
    try:
        return ETREE.XPath(xpath, namespaces=ENERGYML_NAMESPACES)(tree)
    except TypeError:
        return None


def search_element_has_child_xpath(
    tree: ETREE.Element, child_name: str
) -> list:
    """
    Search elements that has a child named (xml tag) as 'child_name'.
    Warning : child_name must contain the namespace (see. ENERGYML_NAMESPACES)
    """
    return list(x for x in energyml_xpath(tree, f"//{child_name}/.."))


def get_uuid(tree: ETREE.Element) -> str:
    _uuids = tree.xpath("@uuid")
    if len(_uuids) <= 0:
        _uuids = tree.xpath("@UUID")
    if len(_uuids) <= 0:
        _uuids = tree.xpath("@Uuid")
    if len(_uuids) <= 0:
        _uuids = tree.xpath("@uid")
    if len(_uuids) <= 0:
        _uuids = tree.xpath("@Uid")
    if len(_uuids) <= 0:
        _uuids = tree.xpath("@UID")
    return _uuids[0]


def get_root_type(tree: ETREE.Element) -> str:
    """Returns the type (xml tag) of the element without the namespace"""
    return tree.xpath("local-name()")


def find_schema_version_in_element(tree: ETREE.ElementTree) -> str:
    """Find the "SchemaVersion" inside an xml content of a energyml file

    :param tree: An energyml xml file content.
    :type tree: bytes

    :returns: The SchemaVersion that contains only the version number. For example, if the xml
        file contains : SchemaVersion="Resqml 2.0.1"
            the result will be : "2.0.1"
    :rtype: str
    """
    _schema_version = tree.xpath("@schemaVersion")
    if _schema_version is None:
        _schema_version = tree.xpath("@SchemaVersion")

    if _schema_version is not None and len(_schema_version) > 0:
        match_version = re.search(r"\d+(\.\d+)*(dev\d+)?", _schema_version[0])
        if match_version is not None:
            return match_version.group(0).replace("dev", "-dev")
    return None



# Copyright (c) 2023-2024 Geosiris.
# SPDX-License-Identifier: Apache-2.0
import inspect
import json
import logging
import random
import re
import sys
import typing
from dataclasses import Field
from enum import Enum
from importlib import import_module
from types import ModuleType
from typing import Any, List, Optional, Union, Dict, Tuple

from .constants import (
    primitives,
    epoch_to_date,
    epoch,
    gen_uuid,
    snake_case,
    pascal_case,
    path_next_attribute,
)
from .manager import (
    get_class_pkg,
    get_class_pkg_version,
    RELATED_MODULES,
    get_related_energyml_modules_name,
    get_sub_classes,
    get_classes_matching_name,
    dict_energyml_modules,
)
from .uri import Uri, parse_uri
from .xml import parse_content_type, ENERGYML_NAMESPACES, parse_qualified_type


def is_enum(cls: Union[type, Any]):
    """
    Returns True if :param:`cls` is an Enum
    :param cls:
    :return:
    """
    if isinstance(cls, type):
        return Enum in cls.__bases__
    return is_enum(type(cls))


def is_primitive(cls: Union[type, Any]) -> bool:
    """
    Returns True if :param:`cls` is a primitiv type or extends Enum
    :param cls:
    :return: bool
    """
    if isinstance(cls, type):
        return cls in primitives or Enum in cls.__bases__
    return is_primitive(type(cls))


def is_abstract(cls: Union[type, Any]) -> bool:
    """
    Returns True if :param:`cls` is an abstract class
    :param cls:
    :return: bool
    """
    if isinstance(cls, type):
        return (
            not is_primitive(cls)
            and (
                cls.__name__.startswith("Abstract")
                or (hasattr(cls, "__dataclass_fields__") and len(cls.__dataclass_fields__)) == 0
            )
            and len(get_class_methods(cls)) == 0
        )
    return is_abstract(type(cls))


def get_module_classes_from_name(mod_name: str) -> List:
    return get_module_classes(sys.modules[mod_name])


def get_module_classes(mod: ModuleType) -> List:
    return inspect.getmembers(mod, inspect.isclass)


def find_class_in_module(module_name, class_name):
    try:
        return getattr(sys.modules[module_name], class_name)
    except:
        for cls_name, cls in get_module_classes_from_name(module_name):
            try:
                if cls_name == class_name or cls.Meta.name == class_name:
                    return cls
            except Exception as e:
                pass
    logging.error(f"Not Found : {module_name}; {class_name}")
    return None


def get_class_methods(cls: Union[type, Any]) -> List[str]:
    """
    Returns the list of the methods names for a specific class.
    :param cls:
    :return:
    """
    return [
        func
        for func in dir(cls)
        if callable(getattr(cls, func)) and not func.startswith("__") and not isinstance(getattr(cls, func), type)
    ]


def get_class_from_name(class_name_and_module: str) -> Optional[type]:
    """
    Return a :class:`type` object matching with the name :param:`class_name_and_module`.
    :param class_name_and_module:
    :return:
    """
    module_name = class_name_and_module[: class_name_and_module.rindex(".")]
    last_ns_part = class_name_and_module[class_name_and_module.rindex(".") + 1 :]
    try:
        # Required to read "CustomData" on eml objects that may contain resqml values
        # ==> we need to import all modules related to the same version of the common
        import_related_module(module_name)
        # return getattr(sys.modules[module_name], last_ns_part)
        return find_class_in_module(module_name, last_ns_part)
    except AttributeError as e:
        # if "2d" in last_ns_part:
        #     logging.debug("replace 2D")
        #     return get_class_from_name(
        #         class_name_and_module.replace("2d", "2D")
        #     )
        # elif "3d" in last_ns_part:
        #     return get_class_from_name(
        #         class_name_and_module.replace("3d", "3D")
        #     )
        # elif last_ns_part[0].islower():
        #     return get_class_from_name(
        #         module_name + "." + last_ns_part[0].upper() + last_ns_part[1:]
        #     )
        # elif "2D" in last_ns_part or "3D" in last_ns_part:
        #     idx = -1
        #     logging.debug(class_name_and_module)
        #     try:
        #         idx = class_name_and_module.rindex("2D") + 2
        #     except:
        #         idx = class_name_and_module.rindex("3D") + 2
        #     if class_name_and_module[idx].isupper():
        #         reformated = (
        #                 class_name_and_module[:idx]
        #                 + class_name_and_module[idx].lower()
        #                 + class_name_and_module[idx + 1:]
        #         )
        #         logging.debug(f"reformated {reformated}")
        #         return get_class_from_name(reformated)
        # else:
        #     logging.debug(e)
        logging.error(e)
    except KeyError:
        logging.error(f"[ERR] module not found : '{module_name}'")
    return None


def get_energyml_module_dev_version(pkg: str, current_version: str):
    accessible_modules = dict_energyml_modules()
    if not current_version.startswith("v"):
        current_version = "v" + current_version

    current_version = current_version.replace("-", "_").replace(".", "_")
    res = []
    if pkg in accessible_modules:
        # logging.debug("\t", pkg, current_version)
        for am_pkg_version in accessible_modules[pkg]:
            if am_pkg_version != current_version and am_pkg_version.startswith(current_version):
                # logging.debug("\t\t", am_pkg_version)
                res.append(get_module_name(pkg, am_pkg_version))

    return res


def get_energyml_class_in_related_dev_pkg(cls: type):
    class_name = cls.__name__
    class_pkg = get_class_pkg(cls)
    class_pkg_version = get_class_pkg_version(cls)

    res = []

    for dev_module_name in get_energyml_module_dev_version(class_pkg, class_pkg_version):
        try:
            res.append(get_class_from_name(f"{dev_module_name}.{class_name}"))
        except Exception as e:
            logging.error(f"FAILED {dev_module_name}.{class_name}")
            logging.error(e)
            pass

    return res


def get_class_from_qualified_type(qualified_type: str) -> Optional[type]:
    return get_class_from_content_type(qualified_type)


def get_class_from_content_type(content_type: str) -> Optional[type]:
    """
    Return a :class:`type` object matching with the content-type :param:`content_type`.
    :param content_type:
    :return:
    """
    ct = None
    try:
        ct = parse_content_type(content_type)
    except AttributeError:
        pass
    if ct is None:
        try:
            ct = parse_qualified_type(content_type)
        except AttributeError:
            pass

    domain = ct.group("domain")
    if domain is None:
        # logging.debug(f"\tdomain {domain} xmlDomain {ct.group('xmlDomain')} ")
        domain = "opc"
    if domain == "opc":
        xml_domain = ct.group("xmlDomain")
        if "." in xml_domain:
            xml_domain = xml_domain[xml_domain.rindex(".") + 1 :]
        # Don't know what to do with http://schemas.f2i-consulting.com/package/2014/metadata/extended-core-properties
        # if "extended" in xml_domain:
        #     xml_domain = xml_domain.replace("extended", "")
        #     if xml_domain.startswith("-"):
        #         xml_domain = xml_domain[1:]
        #
        opc_type = pascal_case(xml_domain).replace("-", "")
        # logging.debug("\tenergyml.opc.opc." + opc_type)
        return get_class_from_name("energyml.opc.opc." + opc_type)
    else:
        domain = ct.group("domain")
        obj_type = ct.group("type")
        if obj_type.lower().startswith("obj_"):  # for resqml201
            # obj_type = "Obj" + obj_type[4:]
            obj_type = obj_type[4:]
        version_num = str(ct.group("domainVersion")).replace(".", "_")
        if "_" not in version_num:
            version_num = re.sub(r"(\d)(\d)", r"\1_\2", version_num)
        if domain.lower() == "resqml" and version_num.startswith("2_0"):
            version_num = "2_0_1"

        # logging.debug(get_module_name(domain, version_num)
        #     + "."
        #     + obj_type)
        return get_class_from_name(get_module_name(domain, version_num) + "." + obj_type)


def get_module_name(domain: str, domain_version: str):
    ns = ENERGYML_NAMESPACES[domain]
    if not domain_version.startswith("v"):
        domain_version = "v" + domain_version
    return f"energyml.{domain}.{domain_version}.{ns[ns.rindex('/') + 1:]}"


def import_related_module(energyml_module_name: str) -> None:
    """
    Import related modules for a specific energyml module. (See. :const:`RELATED_MODULES`)
    :param energyml_module_name:
    :return:
    """
    for related in RELATED_MODULES:
        if energyml_module_name in related:
            for m in related:
                try:
                    import_module(m)
                except Exception as e:
                    pass
                    # logging.error(e)


def list_function_parameters_with_types(func, is_class_function: bool = False) -> Dict[str, Any]:
    """Retrieve parameter names and their types from a function

    Args:
        func (_type_): the function
        is_class_function (bool, optional): If True, remove the first argumen (self or cls). Defaults to False.

    Returns:
        _type_: A dict with parameter name as key and their type as value
    """

    code = func.__code__
    param_names = code.co_varnames[1 : code.co_argcount]  # Get parameter names
    annotations = func.__annotations__  # Get type hints

    # Map parameters to their type hints (if available)
    param_types = {param: annotations.get(param, "Unknown") for param in param_names}

    return param_types


def get_class_fields(cls: Union[type, Any]) -> Dict[str, Field]:
    """
    Return all class fields names, mapped to their :class:`Field` value.
    If a dict is given, this function is the identity
    :param cls:
    :return:
    """
    # for dict object, no change
    if isinstance(cls, dict):
        return cls
    if not isinstance(cls, type):  # if cls is an instance
        cls = type(cls)
    try:
        return cls.__dataclass_fields__
    except AttributeError:
        # print(list_function_parameters_with_types(cls.__new__, True))
        return list_function_parameters_with_types(cls.__new__, True)


def get_class_attributes(cls: Union[type, Any]) -> List[str]:
    """
    returns a list of attributes (not private ones)
    """
    # if not isinstance(cls, type):  # if cls is an instance
    #     cls = type(cls)
    # return list(filter(lambda a: not a.startswith("__"), dir(cls)))
    return list(get_class_fields(cls).keys())


def get_class_attribute_type(cls: Union[type, Any], attribute_name: str):
    fields = get_class_fields(cls)
    try:
        return fields[attribute_name].type
    except IndexError:
        for fn, ft in fields:
            _m = re.match(attribute_name, fn)
            if _m is not None:
                return ft.type

    return None


def get_matching_class_attribute_name(
    cls: Union[type, Any],
    attribute_name: str,
    re_flags=re.IGNORECASE,
) -> Optional[str]:
    """
    From an object and an attribute name, returns the correct attribute name of the class.
    Example : "ObjectVersion" --> object_version.
    This method doesn't only transform to snake case but search into the obj class attributes (or dict keys)
    """
    if isinstance(cls, dict):
        for name in cls.keys():
            if snake_case(name) == snake_case(attribute_name):
                return name
        pattern = re.compile(attribute_name, flags=re_flags)
        for name in cls.keys():
            if pattern.match(name):
                return name
    else:
        class_fields = get_class_fields(cls)

        # a search with the exact value
        for name, cf in class_fields.items():
            if snake_case(name) == snake_case(attribute_name) or (
                "name" in cf.metadata and cf.metadata["name"] == attribute_name
            ):
                return name

        # search regex after to avoid shadowing perfect match
        pattern = re.compile(attribute_name, flags=re_flags)
        for name, cf in class_fields.items():
            # logging.error(f"\t->{name} : {attribute_name} {pattern.match(name)} {('name' in cf.metadata and pattern.match(cf.metadata['name']))}")
            if pattern.match(name) or ("name" in cf.metadata and pattern.match(cf.metadata["name"])):
                return name

    return None


def get_object_attribute(obj: Any, attr_dot_path: str, force_snake_case=True) -> Any:
    """
    returns the value of an attribute given by a dot representation of its path in the object
    example "Citation.Title"

    :param obj:
    :param attr_dot_path:
    :param force_snake_case:
    :return:
    """
    current_attrib_name, path_next = path_next_attribute(attr_dot_path)

    if force_snake_case:
        current_attrib_name = snake_case(current_attrib_name)

    value = None
    if isinstance(obj, list):
        value = obj[int(current_attrib_name)]
    elif isinstance(obj, dict):
        value = obj.get(current_attrib_name, None)
    else:
        try:
            value = getattr(obj, current_attrib_name)
        except AttributeError:
            return None

    # if "." in attr_dot_path:
    #     return get_object_attribute(value, attr_dot_path[len(current_attrib_name) + 1 :])
    # print(f"OLD {obj}\n\t{current_attrib_name}\n\t{path_next}\n\t{value}")
    if path_next is not None:
        return get_object_attribute(value, path_next)
    else:
        return value


def create_default_value_for_type(cls: Any):
    if cls == str:
        return ""
    elif cls == int:
        return 0
    elif cls == float:
        return 0
    elif cls == bool:
        return False
    elif is_enum(cls):
        return cls[cls._member_names_[random.randint(0, len(cls._member_names_) - 1)]]
    elif isinstance(cls, typing.Union.__class__):
        type_list = list(cls.__args__)
        if type(None) in type_list:
            type_list.remove(type(None))  # we don't want to generate none value
        chosen_type = type_list[0]
        return create_default_value_for_type(chosen_type)
    elif cls.__module__ == "typing":
        type_list = list(cls.__args__)
        if type(None) in type_list:
            type_list.remove(type(None))  # we don't want to generate none value

        if cls._name == "List":
            nb_value_for_list = 1
            # On cree une valeur pour ne pas perdre le typage du dessous
            lst = []
            for i in range(nb_value_for_list):
                chosen_type = type_list[0]
                lst.append(create_default_value_for_type(chosen_type))
            return lst
        else:
            chosen_type = type_list[random.randint(0, len(type_list) - 1)]
            return create_default_value_for_type(chosen_type)
    else:
        potential_classes = list(
            filter(
                lambda _c: not is_abstract(_c),
                [cls] + get_sub_classes(cls),
            )
        )
        if len(potential_classes) > 0:
            chosen_type = potential_classes[random.randint(0, len(potential_classes) - 1)]
            if not isinstance(chosen_type, type):
                chosen_type = type(chosen_type)
            return chosen_type()


def get_object_attribute_or_create(
    obj: Any,
    attr_dot_path: str,
    force_snake_case=True,
    fn_create: typing.Callable = lambda o, a_path: create_default_value_for_type(
        get_class_from_simple_name(
            simple_name=get_class_attribute_type(o, a_path),
            energyml_module_context=get_related_energyml_modules_name(o),
        )
    ),
) -> Any:
    """
    returns the value of an attribute given by a dot representation of its path in the object
    example "Citation.Title"

    :param obj:
    :param attr_dot_path:
    :param force_snake_case:
    :return:
    """
    current_attrib_name, path_next = path_next_attribute(attr_dot_path)

    if force_snake_case:
        current_attrib_name = snake_case(current_attrib_name)

    value = None
    try:
        if isinstance(obj, list):
            value = obj[int(current_attrib_name)]
        elif isinstance(obj, dict):
            value = obj.get(current_attrib_name, None)
        else:
            value = getattr(obj, current_attrib_name)
    except Exception:
        pass

    if value is None:
        value = fn_create(obj, current_attrib_name)
        set_attribute_value(obj, current_attrib_name, value)

    if path_next is not None:
        return get_object_attribute(value, path_next)
    else:
        return value


def get_object_attribute_advanced(obj: Any, attr_dot_path: str) -> Any:
    """
    see @get_matching_class_attribute_name and @get_object_attribute
    """
    current_attrib_name = attr_dot_path

    if "." in attr_dot_path:
        current_attrib_name = attr_dot_path.split(".")[0]

    current_attrib_name = get_matching_class_attribute_name(obj, current_attrib_name)

    value = None
    if isinstance(obj, list):
        value = obj[int(current_attrib_name)]
    elif isinstance(obj, dict):
        value = obj[current_attrib_name]
    else:
        value = getattr(obj, current_attrib_name)

    if "." in attr_dot_path:
        return get_object_attribute_advanced(value, attr_dot_path[len(current_attrib_name) + 1 :])
    else:
        return value


def get_object_attribute_no_verif(obj: Any, attr_name: str) -> Any:
    """
    Return the value of the attribute named after param :param:`attr_name` without verification (may raise an exception
    if it doesn't exists).

    Note: attr_name="0" will work if :param:`obj` is of type :class:`List`
    :param obj:
    :param attr_name:
    :return:
    """
    if isinstance(obj, list):
        return obj[int(attr_name)]
    elif isinstance(obj, dict):
        return obj[attr_name]
    else:
        return getattr(obj, attr_name, None)


def get_object_attribute_rgx(obj: Any, attr_dot_path_rgx: str) -> Any:
    """
    see @get_object_attribute. Search the attribute name using regex for values between dots.
    Example : [Cc]itation.[Tt]it\\.*
    """
    current_attrib_name = attr_dot_path_rgx

    attrib_list = re.split(r"(?<!\\)\.+", attr_dot_path_rgx)

    if len(attrib_list) > 0:
        current_attrib_name = attrib_list[0]

    # unescape Dot
    current_attrib_name = current_attrib_name.replace("\\.", ".")

    real_attrib_name = get_matching_class_attribute_name(obj, current_attrib_name)
    if real_attrib_name is not None:
        value = get_object_attribute_no_verif(obj, real_attrib_name)

        if len(attrib_list) > 1:
            return get_object_attribute_rgx(value, attr_dot_path_rgx[len(current_attrib_name) + 1 :])
        else:
            return value
    return None


def get_obj_type(obj: Any) -> str:
    """Return the type name of an object. If obj is already a :class:`type`, return its __name__"""
    if isinstance(obj, type):
        return str(obj.__name__)
    return get_obj_type(type(obj))


def class_match_rgx(
    cls: Union[type, Any],
    rgx: str,
    super_class_search: bool = True,
    re_flags=re.IGNORECASE,
):
    if not isinstance(cls, type):
        cls = type(cls)

    if re.match(rgx, cls.__name__, re_flags):
        return True

    if not is_primitive(cls) and super_class_search:
        for base in cls.__bases__:
            if class_match_rgx(base, rgx, super_class_search, re_flags):
                return True
    return False


def search_attribute_matching_type_with_path(
    obj: Any,
    type_rgx: str,
    re_flags=re.IGNORECASE,
    return_self: bool = True,  # test directly on input object and not only in its attributes
    deep_search: bool = True,  # Search inside a matching object
    super_class_search: bool = True,  # Search inside in super classes of the object
    current_path: str = "",
) -> List[Tuple[str, Any]]:
    """
    Returns a list of tuple (path, value) for each sub attribute with type matching param "type_rgx".
    The path is a dot-version like ".Citation.Title"
    :param obj:
    :param type_rgx:
    :param re_flags:
    :param return_self:
    :param deep_search:
    :param super_class_search:
    :param current_path:
    :return:
    """
    res = []
    if obj is not None:
        if return_self and class_match_rgx(obj, type_rgx, super_class_search, re_flags):
            res.append((current_path, obj))
            if not deep_search:
                return res

    if current_path is None:
        current_path = ""
    if len(current_path) > 0:
        current_path = current_path + "."

    if isinstance(obj, list):
        cpt = 0
        for s_o in obj:
            res = res + search_attribute_matching_type_with_path(
                obj=s_o,
                type_rgx=type_rgx,
                re_flags=re_flags,
                return_self=True,
                deep_search=deep_search,
                current_path=f"{current_path}{cpt}",
                super_class_search=super_class_search,
            )
            cpt = cpt + 1
    elif isinstance(obj, dict):
        for k, s_o in obj.items():
            res = res + search_attribute_matching_type_with_path(
                obj=s_o,
                type_rgx=type_rgx,
                re_flags=re_flags,
                return_self=True,
                deep_search=deep_search,
                current_path=f"{current_path}{k}",
                super_class_search=super_class_search,
            )
    elif not is_primitive(obj):
        for att_name in get_class_attributes(obj):
            res = res + search_attribute_matching_type_with_path(
                obj=get_object_attribute_rgx(obj, att_name),
                type_rgx=type_rgx,
                re_flags=re_flags,
                return_self=True,
                deep_search=deep_search,
                current_path=f"{current_path}{att_name}",
                super_class_search=super_class_search,
            )

    return res


def search_attribute_in_upper_matching_name(
    obj: Any,
    name_rgx: str,
    root_obj: Optional[Any] = None,
    re_flags=re.IGNORECASE,
    current_path: str = "",
) -> Optional[Any]:
    """
    See :func:`search_attribute_matching_name_with_path`. It only returns the value not the path
    :param obj:
    :param name_rgx:
    :param root_obj:
    :param re_flags:
    :param current_path:
    :return:
    """
    # print(f"Searching {name_rgx} in {obj} : {root_obj}")
    elt_list = search_attribute_matching_name(obj, name_rgx, search_in_sub_obj=False, deep_search=False)
    if elt_list is not None and len(elt_list) > 0:
        return elt_list

    if len(current_path) != 0:  # obj != root_obj:
        upper_path = current_path[: current_path.rindex(".")]
        # print(f"\t {upper_path} ")
        if len(upper_path) > 0:
            return search_attribute_in_upper_matching_name(
                obj=get_object_attribute(root_obj, upper_path),
                name_rgx=name_rgx,
                root_obj=root_obj,
                re_flags=re_flags,
                current_path=upper_path,
            )
        else:
            return search_attribute_in_upper_matching_name(
                obj=root_obj,
                name_rgx=name_rgx,
                root_obj=root_obj,
                re_flags=re_flags,
                current_path=upper_path,
            )

    return None


def search_attribute_matching_type(
    obj: Any,
    type_rgx: str,
    re_flags=re.IGNORECASE,
    return_self: bool = True,  # test directly on input object and not only in its attributes
    deep_search: bool = True,  # Search inside a matching object
    super_class_search: bool = True,  # Search inside in super classes of the object
) -> List[Any]:
    """
    See :func:`search_attribute_matching_type_with_path`. It only returns the value not the path
    :param obj:
    :param type_rgx:
    :param re_flags:
    :param return_self:
    :param deep_search:
    :param super_class_search:
    :return:
    """
    return [
        val
        for path, val in search_attribute_matching_type_with_path(
            obj=obj,
            type_rgx=type_rgx,
            re_flags=re_flags,
            return_self=return_self,
            deep_search=deep_search,
            super_class_search=super_class_search,
        )
    ]


def search_attribute_matching_name_with_path(
    obj: Any,
    name_rgx: str,
    re_flags=re.IGNORECASE,
    current_path: str = "",
    deep_search: bool = True,  # Search inside a matching object
    search_in_sub_obj: bool = True,  # Search in obj attributes
) -> List[Tuple[str, Any]]:
    """
    Returns a list of tuple (path, value) for each sub attribute with type matching param "name_rgx".
        The path is a dot-version like ".Citation.Title"
    :param obj:
    :param name_rgx:
    :param re_flags:
    :param current_path:
    :param deep_search:
    :param search_in_sub_obj:
    :return:
    """
    # while name_rgx.startswith("."):
    #     name_rgx = name_rgx[1:]
    # current_match = name_rgx
    # next_match = current_match
    # if "." in current_match:
    #     attrib_list = re.split(r"(?<!\\)\.+", name_rgx)
    #     current_match = attrib_list[0]
    #     next_match = ".".join(attrib_list[1:])
    current_match, next_match = path_next_attribute(name_rgx)
    res = []

    if current_path is None:
        current_path = ""
    if len(current_path) > 0:
        current_path = current_path + "."

    match_path_and_obj = []
    not_match_path_and_obj = []
    if isinstance(obj, list):
        cpt = 0
        for s_o in obj:
            match = re.match(current_match.replace("\\.", "."), str(cpt), flags=re_flags)
            if match is not None:
                match_value = match.group(0)
                match_path_and_obj.append((f"{current_path}{cpt}", s_o))
            else:
                not_match_path_and_obj.append((f"{current_path}{cpt}", s_o))
            cpt = cpt + 1
    elif isinstance(obj, dict):
        for k, s_o in obj.items():
            match = re.match(current_match.replace("\\.", "."), k, flags=re_flags)
            if match is not None:
                match_value = match.group(0)
                match_path_and_obj.append((f"{current_path}{match_value}", s_o))
            else:
                not_match_path_and_obj.append((f"{current_path}{k}", s_o))
    elif not is_primitive(obj):
        match_value = get_matching_class_attribute_name(obj, current_match.replace("\\.", "."))
        if match_value is not None:
            match_path_and_obj.append(
                (
                    f"{current_path}{match_value}",
                    get_object_attribute_no_verif(obj, match_value),
                )
            )
        for att_name in get_class_attributes(obj):
            if att_name != match_value:
                not_match_path_and_obj.append(
                    (
                        f"{current_path}{att_name}",
                        get_object_attribute_no_verif(obj, att_name),
                    )
                )

    for matched_path, matched in match_path_and_obj:
        if next_match is not None:  # next_match is different, match is not final
            res = res + search_attribute_matching_name_with_path(
                obj=matched,
                name_rgx=next_match,
                re_flags=re_flags,
                current_path=matched_path,
                deep_search=False,  # no deep with partial
                search_in_sub_obj=False,  # no partial search in sub obj with no match
            )
        else:  # a complete match
            res.append((matched_path, matched))
            if deep_search:
                res = res + search_attribute_matching_name_with_path(
                    obj=matched,
                    name_rgx=name_rgx,
                    re_flags=re_flags,
                    current_path=matched_path,
                    deep_search=deep_search,  # no deep with partial
                    search_in_sub_obj=True,
                )
    if search_in_sub_obj:
        for not_matched_path, not_matched in not_match_path_and_obj:
            res = res + search_attribute_matching_name_with_path(
                obj=not_matched,
                name_rgx=name_rgx,
                re_flags=re_flags,
                current_path=not_matched_path,
                deep_search=deep_search,
                search_in_sub_obj=True,
            )

    return res


def search_attribute_matching_name(
    obj: Any,
    name_rgx: str,
    re_flags=re.IGNORECASE,
    deep_search: bool = True,  # Search inside a matching object
    search_in_sub_obj: bool = True,  # Search in obj attributes
) -> List[Any]:
    """
    See :func:`search_attribute_matching_name_with_path`. It only returns the value not the path

    :param obj:
    :param name_rgx:
    :param re_flags:
    :param deep_search:
    :param search_in_sub_obj:
    :return:
    """
    return [
        val
        for path, val in search_attribute_matching_name_with_path(
            obj=obj,
            name_rgx=name_rgx,
            re_flags=re_flags,
            deep_search=deep_search,
            search_in_sub_obj=search_in_sub_obj,
        )
    ]


def set_attribute_from_json_str(obj: Any, json_input: str) -> None:
    set_attribute_from_dict(obj=obj, values=json.loads(json_input))


def set_attribute_from_dict(obj: Any, values: Dict) -> None:
    for k, v in values.items():
        if isinstance(v, dict):
            set_attribute_from_dict(obj=get_object_attribute(obj=obj, attr_dot_path=k), values=v)
        elif isinstance(v, list):
            obj_list: List = get_object_attribute(obj=obj, attr_dot_path=k)
            while len(obj_list) > len(v):
                obj_list.pop()
            for i in range(len(v)):
                set_attribute_from_dict(obj=get_object_attribute(obj=obj_list, attr_dot_path="i"), values=v[i])
        else:
            set_attribute_from_path(obj=obj, attribute_path=k, value=v)


def set_attribute_from_path(obj: Any, attribute_path: str, value: Any):
    """
    Changes the value of a (sub)attribute.
    Example :
        data = {
            "a": {
                "b": [ "v_x", { "c": "v_old" } ]
            }
        }
        set_attribute_from_path(data, "a.b.1.c", "v_new")

        # result is :

        data == {
            "a": {
                "b": [ "v_x", { "c": "v_new" } ]
            }
        }

    :param obj:
    :param attribute_path:
    :param value:
    :return:
    """
    upper = obj
    current_attrib_name, path_next = path_next_attribute(attribute_path)
    if path_next is not None:
        set_attribute_from_path(
            get_object_attribute(
                obj,
                current_attrib_name,
            ),
            path_next,
            value,
        )
    else:
        current_attrib_real_name = get_matching_class_attribute_name(upper, current_attrib_name)
        created = False
        if current_attrib_real_name is not None:
            attrib_class = get_obj_attribute_class(upper, current_attrib_real_name)
            if attrib_class is not None and is_enum(attrib_class):
                created = True
                val_snake = snake_case(value)
                setattr(
                    upper,
                    current_attrib_real_name,
                    list(
                        filter(
                            lambda ev: snake_case(ev) == val_snake,
                            attrib_class._member_names_,
                        )
                    )[0],
                )
        if not created:  # If previous test failed, the attribute did not exist in the object, we create it
            if isinstance(upper, dict):
                upper[current_attrib_name] = value
            else:
                setattr(upper, current_attrib_name, value)


def set_attribute_value(obj: any, attribute_name_rgx, value: Any):
    copy_attributes(obj_in={attribute_name_rgx: value}, obj_out=obj, ignore_case=True)


def copy_attributes(
    obj_in: any,
    obj_out: Any,
    only_existing_attributes: bool = True,
    ignore_case: bool = True,
):
    in_att_list = get_class_attributes(obj_in)
    for k_in in in_att_list:
        p_list = search_attribute_matching_name_with_path(
            obj=obj_out,
            name_rgx=k_in,
            re_flags=re.IGNORECASE if ignore_case else 0,
            deep_search=False,
            search_in_sub_obj=False,
        )
        path = None
        if p_list is not None and len(p_list) > 0:
            path, _ = p_list[0]
        if not only_existing_attributes or path is not None:
            set_attribute_from_path(
                obj_out,
                path or k_in,
                get_object_attribute(obj_in, k_in, False),
            )


# Utility functions


def get_obj_uuid(obj: Any) -> str:
    """
    Return the object uuid (attribute must match the following regex : "[Uu]u?id|UUID").
    :param obj:
    :return:
    """
    return get_object_attribute_rgx(obj, "[Uu]u?id|UUID")


def get_obj_version(obj: Any) -> str:
    """
    Return the object version (check for "object_version" or "version_string" attribute).
    :param obj:
    :return:
    """
    try:
        return get_object_attribute_no_verif(obj, "object_version")
    except AttributeError as e:
        try:
            return get_object_attribute_no_verif(obj, "version_string")
        except Exception:
            logging.error(f"Error with {type(obj)}")
            # raise e


def get_obj_pkg_pkgv_type_uuid_version(
    obj: Any,
) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]]:
    """
    return from an energyml object or a DOR a tuple :
        - package : e.g. resqml|eml|witsml|prodml
        - package version : e.g. 20
        - type : e.g. obj_TriangulatedSetRepresentation
        - uuid
        - object version
    :param obj:
    :return:
    """
    pkg: Optional[str] = get_class_pkg(obj)
    pkg_v: Optional[str] = get_class_pkg_version(obj)
    obj_type: Optional[str] = get_object_type_for_file_path_from_class(obj)
    obj_uuid = get_obj_uuid(obj)
    obj_version = get_obj_version(obj)

    ct = None
    try:
        ct = get_object_attribute_no_verif(obj, "content_type")
    except:
        pass

    if ct is not None:
        ct_match = parse_content_type(ct)
        logging.debug("ct : ", ct_match)
        if ct_match is not None:
            pkg = ct_match.group("domain")
            pkg_v = ct_match.group("domainVersion")
            obj_type = ct_match.group("type")
    else:
        try:
            qt = get_object_attribute_no_verif(obj, "qualified_type")
            qt_match = parse_qualified_type(qt)
            logging.debug("qt : ", qt, obj.__dict__, qt_match)
            if qt_match is not None:
                pkg = qt_match.group("domain")
                pkg_v = qt_match.group("domainVersion")
                obj_type = qt_match.group("type")
        except:
            pass

    # flattening version
    if pkg_v is not None:
        pkg_v = pkg_v.replace(".", "")

    return pkg, pkg_v, obj_type, obj_uuid, obj_version


def get_obj_identifier(obj: Any) -> str:
    """
    Generates an objet identifier as : 'OBJ_UUID.OBJ_VERSION'
    If the object version is None, the result is 'OBJ_UUID.'
    :param obj:
    :return: str
    """
    obj_obj_version = get_obj_version(obj)
    if obj_obj_version is None:
        obj_obj_version = ""
    obj_uuid = get_obj_uuid(obj)
    return f"{obj_uuid}.{obj_obj_version}"


def get_obj_uri(obj: Any, dataspace: Optional[str] = None) -> Uri:
    """
    Generates an objet etp Uri from an objet or a DOR
    :param obj:
    :param dataspace: the etp dataspace
    :return: str
    """
    (
        domain,
        domain_version,
        object_type,
        obj_uuid,
        obj_version,
    ) = get_obj_pkg_pkgv_type_uuid_version(obj)

    return Uri(
        dataspace=dataspace,
        domain=domain,
        domain_version=domain_version,
        object_type=object_type,
        uuid=obj_uuid,
        version=obj_version,
    )


def get_direct_dor_list(obj: Any) -> List[Any]:
    """
    Search all sub attribute of type "DataObjectreference".
    :param obj:
    :return:
    """
    return search_attribute_matching_type(obj, "DataObjectreference")


def get_data_object_type(cls: Union[type, Any], print_dev_version=True, nb_max_version_digits=2):
    return get_class_pkg(cls) + "." + get_class_pkg_version(cls, print_dev_version, nb_max_version_digits)


def get_qualified_type_from_class(cls: Union[type, Any], print_dev_version=True):
    return (
        get_data_object_type(cls, print_dev_version, 2).replace(".", "")
        + "."
        + get_object_type_for_file_path_from_class(cls)
    )


def get_object_uri(obj: any, dataspace: Optional[str] = None) -> Optional[Uri]:
    """Returns an ETP URI"""
    return parse_uri(f"eml:///dataspace('{dataspace or ''}')/{get_qualified_type_from_class(obj)}({get_obj_uuid(obj)})")


def dor_to_uris(dor: Any, dataspace: Optional[str] = None) -> Optional[Uri]:
    """
    Transform a DOR into an etp uri
    """
    result = None
    try:
        value = get_object_attribute_no_verif(dor, "qualified_type")
        result = parse_qualified_type(value)
    except Exception as e:
        print(e)
        try:
            value = get_object_attribute_no_verif(dor, "content_type")
            result = parse_content_type(value)
        except Exception as e2:
            print(e2)

    if result is None:
        return None

    return Uri(
        dataspace=dataspace,
        domain=result.group("domain"),
        domain_version=result.group("domainVersion"),
        object_type=result.group("type"),
        uuid=dor.uuid,
        version=dor.object_version,
    )


def get_content_type_from_class(cls: Union[type, Any], print_dev_version=True, nb_max_version_digits=2):
    if not isinstance(cls, type):
        cls = type(cls)

    if ".opc." in cls.__module__:
        if cls.__name__.lower() == "coreproperties":
            return "application/vnd.openxmlformats-package.core-properties+xml"
    else:
        return (
            "application/x-"
            + get_class_pkg(cls)
            + "+xml;version="
            + get_class_pkg_version(cls, print_dev_version, nb_max_version_digits)
            + ";type="
            + get_object_type_for_file_path_from_class(cls)
        )

    logging.error(f"@get_content_type_from_class not supported type : {cls}")
    return None


def get_object_type_for_file_path_from_class(cls) -> str:
    if not isinstance(cls, type):
        cls = type(cls)
    classic_type = get_obj_type(cls)

    for parent_cls in cls.__bases__:
        try:
            if (
                classic_type.lower() in parent_cls.Meta.name.lower()
            ):  # to work with 3d transformed in 3D and Obj[A-Z] in obj_[A-Z]
                return parent_cls.Meta.name
        except AttributeError:
            pass

    return classic_type


def get_obj_attribute_class(
    cls: Any,
    attribute_name: Optional[str],
    random_for_typing: Optional[bool] = False,
    no_abstract: Optional[bool] = True,
):
    """
    Return an instantiable class for an attribute of :param cls:.
    If the attribute is defined with typing with multiple possibility (like tuple or union), the first one
    is selected or a random one (depending on the value of the :param random_for_typing:)
    :param cls:
    :param attribute_name:
    :param random_for_typing:
    :param no_abstract: if True, the returned typed will not be an abstract class
    :return:
    """
    chosen_type = None
    if cls is not None and attribute_name is not None:
        if isinstance(cls, dict):
            attribute_name_real = get_matching_class_attribute_name(cls, attribute_name)
            if attribute_name_real is not None:
                return type(cls[attribute_name_real])
        elif not isinstance(cls, type) and cls.__module__ != "typing":
            return get_obj_attribute_class(type(cls), attribute_name, random_for_typing)
        elif cls.__module__ == "typing":
            type_list = list(cls.__args__)
            if type(None) in type_list:
                type_list.remove(type(None))  # we don't want to generate none value

            if random_for_typing:
                chosen_type = type_list[random.randint(0, len(type_list) - 1)]
            else:
                chosen_type = type_list[0]
            return get_obj_attribute_class(chosen_type, None, random_for_typing)
        else:
            if attribute_name is not None and len(attribute_name) > 0:
                cls = get_class_from_simple_name(
                    simple_name=get_class_fields(cls)[attribute_name].type,
                    energyml_module_context=get_related_energyml_modules_name(cls),
                )
            potential_classes = [cls] + get_sub_classes(cls)
            if no_abstract:
                potential_classes = list(filter(lambda _c: not is_abstract(_c), potential_classes))
            if random_for_typing:
                chosen_type = potential_classes[random.randint(0, len(potential_classes) - 1)]
            else:
                chosen_type = potential_classes[0]
            #             print(f"chosen_type {chosen_type}")

            if cls.__module__ == "typing":
                return get_obj_attribute_class(chosen_type, None, random_for_typing)

    elif cls is not None:
        if isinstance(cls, typing.Union.__class__):
            type_list = list(cls.__args__)
            if type(None) in type_list:
                type_list.remove(type(None))  # we don't want to generate none value
            chosen_type = type_list[random.randint(0, len(type_list))]
        elif cls.__module__ == "typing":
            type_list = list(cls.__args__)
            if type(None) in type_list:
                type_list.remove(type(None))  # we don't want to generate none value

            if cls._name == "List":
                nb_value_for_list = random.randint(2, 3)
                lst = []
                for i in range(nb_value_for_list):
                    chosen_type = type_list[random.randint(0, len(type_list) - 1)]
                    lst.append(
                        _random_value_from_class(
                            chosen_type,
                            get_related_energyml_modules_name(cls),
                            attribute_name,
                            list,
                        )
                    )
                return lst
            else:
                chosen_type = type_list[random.randint(0, len(type_list) - 1)]

    return chosen_type


#  RANDOM


def get_class_from_simple_name(simple_name: str, energyml_module_context=None) -> type:
    """
    Search for a :class:`type` depending on the simple class name :param:`simple_name`.
    :param simple_name:
    :param energyml_module_context:
    :return:
    """
    if energyml_module_context is None:
        energyml_module_context = []
    try:
        return eval(simple_name)
    except NameError as e:
        for mod in energyml_module_context:
            try:
                exec(f"from {mod} import *")
                # required to be able to access to type in
                # typing values like "List[ObjectAlias]"
            except ModuleNotFoundError:
                pass
        return eval(simple_name)


def _gen_str_from_attribute_name(attribute_name: Optional[str], _parent_class: Optional[type] = None) -> str:
    """
    Generate a str from the attribute name. The result is not the same for an attribute named "Uuid" than for an
    attribute named "mime_type" for example.
    :param attribute_name:
    :param _parent_class:
    :return:
    """
    attribute_name_lw = attribute_name.lower()
    if attribute_name is not None:
        if attribute_name_lw == "uuid" or attribute_name_lw == "uid":
            return gen_uuid()
        elif attribute_name_lw == "title":
            return f"{_parent_class.__name__} title (" + str(random_value_from_class(int)) + ")"
        elif attribute_name_lw == "schema_version" and get_class_pkg_version(_parent_class) is not None:
            return get_class_pkg_version(_parent_class)
        elif re.match(r"\w*version$", attribute_name_lw):
            return str(random_value_from_class(int))
        elif re.match(r"\w*date_.*", attribute_name_lw):
            return epoch_to_date(epoch())
        elif re.match(r"path_in_.*", attribute_name_lw):
            return f"/FOLDER/{gen_uuid()}/a_patch{random.randint(0, 30)}"
        elif "mime_type" in attribute_name_lw and (
            "external" in _parent_class.__name__.lower() and "part" in _parent_class.__name__.lower()
        ):
            return f"application/x-hdf5"
        elif "type" in attribute_name_lw:
            if attribute_name_lw.startswith("qualified"):
                return get_qualified_type_from_class(get_classes_matching_name(_parent_class, "Abstract")[0])
            if attribute_name_lw.startswith("content"):
                return get_content_type_from_class(get_classes_matching_name(_parent_class, "Abstract")[0])
    return (
        "A random str "
        + (f"[{attribute_name}] " if attribute_name is not None else "")
        + "("
        + str(random_value_from_class(int))
        + ")"
    )


def random_value_from_class(cls: type):
    """
    Generate a random value for a :class:`type`. All attributes should be filled with random values.
    :param cls:
    :return:
    """
    energyml_module_context = []
    if not is_primitive(cls):
        # import_related_module(cls.__module__)
        energyml_module_context = get_related_energyml_modules_name(cls)
    return _random_value_from_class(
        cls=cls,
        energyml_module_context=energyml_module_context,
        attribute_name=None,
    )


def _random_value_from_class(
    cls: Any,
    energyml_module_context: List[str],
    attribute_name: Optional[str] = None,
    _parent_class: Optional[type] = None,
):
    """
    Generate a random value for a :class:`type`. All attributes should be filled with random values.
    :param cls:
    :param energyml_module_context:
    :param attribute_name:
    :param _parent_class: the :class:`type`of the parent object
    :return:
    """

    try:
        if isinstance(cls, str) or cls == str:
            return _gen_str_from_attribute_name(attribute_name, _parent_class)
        elif isinstance(cls, int) or cls == int:
            return random.randint(0, 10000)
        elif isinstance(cls, float) or cls == float:
            return random.randint(0, 1000000) / 100.0
        elif isinstance(cls, bool) or cls == bool:
            return random.randint(0, 1) == 1
        elif is_enum(cls):
            return cls[cls._member_names_[random.randint(0, len(cls._member_names_) - 1)]]
        elif isinstance(cls, typing.Union.__class__):
            type_list = list(cls.__args__)
            if type(None) in type_list:
                type_list.remove(type(None))  # we don't want to generate none value
            chosen_type = type_list[random.randint(0, len(type_list))]
            return _random_value_from_class(chosen_type, energyml_module_context, attribute_name, cls)
        elif cls.__module__ == "typing":
            type_list = list(cls.__args__)
            if type(None) in type_list:
                type_list.remove(type(None))  # we don't want to generate none value

            if cls._name == "List":
                nb_value_for_list = random.randint(2, 3)
                lst = []
                for i in range(nb_value_for_list):
                    chosen_type = type_list[random.randint(0, len(type_list) - 1)]
                    lst.append(
                        _random_value_from_class(
                            chosen_type,
                            energyml_module_context,
                            attribute_name,
                            list,
                        )
                    )
                return lst
            else:
                chosen_type = type_list[random.randint(0, len(type_list) - 1)]
                return _random_value_from_class(
                    chosen_type,
                    energyml_module_context,
                    attribute_name,
                    _parent_class,
                )
        else:
            potential_classes = list(
                filter(
                    lambda _c: not is_abstract(_c),
                    [cls] + get_sub_classes(cls),
                )
            )
            if len(potential_classes) > 0:
                chosen_type = potential_classes[random.randint(0, len(potential_classes) - 1)]
                args = {}
                for k, v in get_class_fields(chosen_type).items():
                    # logging.debug(f"get_class_fields {k} : {v}, { isinstance(v, type)}, {v}")
                    args[k] = _random_value_from_class(
                        cls=(
                            get_class_from_simple_name(
                                simple_name=getattr(v, "type", v),
                                energyml_module_context=energyml_module_context,
                            )
                            if not isinstance(v, type) and not v.__module__ == "typing"
                            else v
                        ),
                        energyml_module_context=energyml_module_context,
                        attribute_name=k,
                        _parent_class=chosen_type,
                    )

                if not isinstance(chosen_type, type):
                    chosen_type = type(chosen_type)

                return chosen_type(**args)

    except Exception as e:
        logging.error(f"exception on attribute '{attribute_name}' for class {cls} :")
        raise e

    logging.error(f"@_random_value_from_class Not supported object type generation {cls}")
    return None
